// Version 1.2

// configTester.cpp : Defines the entry point for the console application.
// Written By: Patrick Mattingly
// Email: pat@iads-soft.com
// Google Group: http://groups.google.com/group/iads
//
// Upgrades:
//
//   Version 1.3
//     1. Added new Collections for the Testpoint, Selections and Thresholds logs

//   Version 1.2
//     1. Added support for config file versions 3.4.5, 4.0 and 5.0
// 
//   Version 1.1
//     1. Added Binary tree capability for efficient parameter replacement
//     2. Added capability to replace the "||" string in a derived equation with char 255
//        to work properly in IADS client
//     3. Now works with 4.0 configuration files
//
//   Version 1.0
//     1. Initital Version
//
// *****************************************************
//
//  Making General purpose IADS configuration queries
//
// *****************************************************
//
//   Query string construction has the following form:
//     keyword <field name> from <table name> where <qualifiers>
//
//     Keywords are:
//       1. Select - Selected values are returned in BSTR array
//       2. Update - Query user passes in a string to update in the configuration file.
//       3. Delete - Deletes one or more rows in a configuration file
//       4. Create - Create a table in the configuration file.
//
//     field names:
//       These are directly from the configuration file, therefore the query user must have knowledge of it's
//       construction.  Multiple field names are separated by commas.  A wild card of '*' can be used in
//       which case the entire row is returned with individual field values delimited by the '|' vertical
//       pipe character.
//
//     table name:
//       Table names are those that are in the configuration file, therefore the query user must have
//       knowledge of it's constrcution.
//
//     qualifiers:
//       Qualifiers take the form of: field name = value.  All values of type string must be enclosed in
//       single quotes, example: Group = 'Loads'.
//
//    Query Examples:  
//
//      1. "Select * from Desktops"
//           - Get all fields for all entries in a particular table.
//      2. "Select * from Desktops where Group = 'Flutter'"
//           - Get all fields for selected entries using a "where" clause.
//      3. "Select * from Desktops where Group = 'Flutter' && AnalysisWindowName = 'DoubleIntTest'"
//           - Get all fields for selected entries using a compound "where" clause.
//      4. "Select SubGroup from Desktops where Group = 'Flutter' && AnalysisWindowName = 'DoubleIntTest'
//           - Get one field for selected entries using a compound "where" clause.
//      5. "Select System.RowNumber from Desktops"
//            - Get one field for selected entries using a compound "where" clause.  Note usage of "System.RowNumber
//              which selects a row based on it's unique Id (be careful with this).
//      6. "Select System.TableDefinition from ParameterAlias"
//            - Get the table definition string
//            - Check if a table is present
//      7. "Update BogusDesktops set * = a|b|c|d|e"
//            - Modify all fields in all rows in a table
//      8. "Update BogusDesktops set * = a|b|c|d|e where Group = 'Flutter'"
//            - Modify all fields in the specified row(s) based on a match in a single field.
//      9. "Update BogusDesktops set SubGroup = 'Pat' where Group = 'FQ'"
//            - Modify a single field in a specified row.
//     10. "Create table BogusTable (Group String, X Int, Y Int, AutoScale list(True, False), Classification list(high,medium,low) )"
//            - Creates a table called BogusTable with fields as shown above.  Note usage of the list type.  This
//              will create a drop down list in the configuration tool for easier user entry.
//     11. "Delete * from BogusDesktops where Group = 'Flutter'"
//            - Delete every row from the BogusDesktopns table where the field value is Flutter
//     12. "Delete * from BogusDesktops where System.RowNumber = 3"
//            - Delete using "built-in" unique system id or row number

#include "stdafx.h"
#include "conio.h"
#include <atlbase.h>

#include "IadsConfigInterface.h"
#include "IadsConfigInterface_i.c"

//
// This #define section is used to control which tests get executed
//

//#define DO_EVENTS_TEST

//#define DO_THRESHOLDS_TEST
//#define DO_TESTPOINTS_TEST
//#define DO_PLANNEDTPS_TEST
//#define DO_SELECTIONS_TEST

#define DO_SELECT_QUERY_TEST
//#define DO_UPDATE_QUERY_TEST
//#define DO_CREATE_QUERY_TEST
//#define DO_CREATE_CONFIGS_TEST
//#define DO_DELETE_QUERY_TEST
//#define DO_PARAMETER_DEFAULTS_TEST
//#define DO_LARGE_PARAMETER_DEFAULTS_TEST
//#define MERGE_PARAMETERS_FROM_TWO_CONFIGS

const int cMaxString = 1000 ;

// Globals
IIadsConfig* gConfig = NULL ;
IParameterDefaults* gDefaults = NULL ;
IParameterSets* gSets = NULL ;
IEvents* gEvents = NULL ;
IThresholds* gThresholds = NULL ;
ITestpoints* gTestpoints = NULL ;
ISelections* gSelections = NULL ;
IPlannedTestPoints* gPlannedTps = NULL ;

bool gDefaultsChanged = false ;
bool gSetsChanged = false ;
char tmpBuf1[cMaxString+1] ;
char tmpBuf2[cMaxString+1] ;
char tmpBuf3[cMaxString+1] ;
char tmpBuf4[cMaxString+1] ;
char tmpBuf5[cMaxString+1] ;
char tmpBuf6[cMaxString+1] ;
char tmpBuf7[cMaxString+1] ;
char tmpBuf8[cMaxString+1] ;
char tmpBuf9[cMaxString+1] ;
char tmpBuf10[cMaxString+1] ;
BSTR b1, b2, b3, b4, b5, b6, b7, b8, b9, b10 ;

// Forward Declarations
void GetCollectionInterfaces(void) ;
void OpenMsgLog(void) ;
void ShowVersion(void) ;
void ShowVersionFromFile(void) ;

void ShowAllSets(void) ;
void AddSets(void) ;

void ShowAllParameters(void) ;
void AddParameters(void) ;
void AddLargeParameters(int maxParms) ;
void SearchLargeParameters(void) ;
void RemoveParameters(void) ;
void MergeParametersFromTwoConfigs(void) ;

void ShowAllEvents(void) ;
void AddEvents(void) ;

void ShowAllThresholds(void) ;
//void AddThresholds(void) ;

void ShowAllTestpoints(void) ;
void AddTestpoints(void) ;

void ShowAllSelections(void) ;
//void AddSelections(void)

//void ShowAllPlannedTestpoints(void) ;
//void AddPlannedTestpoints(void)

void SelectQueryExamples(void) ;
void UpdateQueryExamples(void) ;
void CreateQueryExamples(void) ;
void DeleteQueryExamples(void) ;
void PrintQueryResults( VARIANT* v ) ;


void shutdown(void)
{
  if( gSets != NULL ) gSets->Release() ;
  if( gDefaults != NULL ) gDefaults->Release() ;
  if( gConfig != NULL ) gConfig->Release() ;

	printf("\n\nFinished...Hit any key to continue\n");

  while(!_kbhit()) ;

  ::CoUninitialize() ;
	exit( 0 );
}

//
// Main Program
//
void main(int argc, char* argv[])
{
  ::CoInitialize(NULL) ;

  // Create the IADS config object
  HRESULT hr = ::CoCreateInstance( CLSID_IadsConfig, NULL, CLSCTX_ALL, IID_IIadsConfig, (void**)&gConfig ) ;
  if( hr != S_OK || gConfig == NULL )
  {
    printf("** Unable to create the IadsConfig object ?\n" ) ;
    shutdown() ;
  }

//
// Show Verson direcly by reading the file usefull if you just want to check 
// versions without opening the the file via the interface
//
  ShowVersionFromFile() ;
  OpenMsgLog() ;


//
// Open an existing config file
//
  printf("\n*\n* Open the existing configuration file\n*\n" ) ;
  hr = gConfig->Open(L"pfConfig5.0", VARIANT_TRUE ) ;
  //hr = gConfig->Open(L"pfConfig4.0", VARIANT_TRUE ) ;

#ifdef MERGE_PARAMETERS_FROM_TWO_CONFIGS

  MergeParametersFromTwoConfigs() ;

#endif

  /*
  hr = gConfig->Open(L"\\Projects\\IADS\\Com\\DataTools\\IadsConfigInterface\\TestPrograms\\CPP\\pfConfig", VARIANT_TRUE ) ;
  if( hr != S_OK )
  {
    printf("** Unable to open configuration file\n" ) ;
    shutdown() ;
  }

  ShowVersion() ;
  */

//*****************************************************************************
//
//                                 Events Test
//
//*****************************************************************************
#ifdef DO_EVENTS_TEST

  // Get the parameter Sets collection
  hr = gConfig->get_Events(&gEvents) ; // Make sure you release this before releasing the config object
  if( hr != S_OK || gEvents == NULL )
  {
    printf("** Error: Getting the Events collection\n") ;
    shutdown() ;
  }

  ShowAllEvents() ;
  AddEvents() ;

  gEvents->SaveTable() ;

  ShowAllEvents() ;

//  RemoveEvents() ;
//  ShowAllEvents() ;

#endif


//*****************************************************************************
//
//                                 Thresholds Test
//
//*****************************************************************************
#ifdef DO_THRESHOLDS_TEST

  // Get the parameter Sets collection
  hr = gConfig->get_Thresholds(&gThresholds) ; // Make sure you release this before releasing the config object
  if( hr != S_OK || gThresholds == NULL )
  {
    printf("** Error: Getting the Thresholds collection\n") ;
    shutdown() ;
  }

  ShowAllThresholds() ;
  //AddThresholdss() ;
  //gThresholds->SaveTable() ;

  //ShowAllThresholds() ;

//  RemoveThresholds() ;
//  ShowAllThresholds() ;

#endif


//*****************************************************************************
//
//                                 TestPoints Test
//
//*****************************************************************************
#ifdef DO_TESTPOINTS_TEST

  // Get the Test Point collection
  hr = gConfig->get_Testpoints(&gTestpoints) ; // Make sure you release this before releasing the config object
  if( hr != S_OK || gTestpoints == NULL )
  {
    printf("** Error: Getting the Thresholds collection\n") ;
    shutdown() ;
  }

  ShowAllTestpoints() ;
  AddTestpoints() ;
  gTestpoints->SaveTable() ;

  ShowAllTestpoints() ;

//  RemoveThresholds() ;
//  ShowAllThresholds() ;

#endif


//*****************************************************************************
//
//                               Query Tests
//
//*****************************************************************************

// Note: All Queries must be done MUST be done after an Open for Create
// Query keywords are: select, udpate, delete, create
// See routines below for test setup and expected results.
//
#ifdef DO_SELECT_QUERY_TEST

  SelectQueryExamples() ;

#endif

#ifdef DO_UPDATE_QUERY_TEST

  // Dump the Desktops for starters - must be done after an Open for Create
  UpdateQueryExamples() ;

#endif

#ifdef DO_CREATE_QUERY_TEST

  // Dump the Desktops for starters - must be done after an Open for Create
  CreateQueryExamples() ;

#endif

#ifdef DO_DELETE_QUERY_TEST

  // Dump the Desktops for starters - must be done after an Open for Create
  DeleteQueryExamples() ;

#endif


//*****************************************************************************
//
//                           Parameter Defaults Test
//
//*****************************************************************************
#ifdef DO_PARAMETER_DEFAULTS_TEST

  GetCollectionInterfaces() ;
  ShowVersion() ;

//
// Parameter Set Manipulation
//
  ShowAllSets() ;
  AddSets() ;
  ShowAllSets() ;

//
// Parameter Defaults Manipulation
//
  ShowAllParameters() ;
  AddParameters() ;
  ShowAllParameters() ;
  RemoveParameters() ;
 // gDefaults->RemoveAll() ; gDefaultsChanged = true ;

//
// Save Tables if they were modified
//
  if( gDefaultsChanged == true ) gDefaults->SaveTable() ;
  if( gSetsChanged == true ) gSets->SaveTable() ;

  hr = gConfig->Close(VARIANT_TRUE) ;
  if( hr != S_OK )
  {
    printf("** Close and Save failed...\n" ) ;
    gConfig = NULL ;
    gSets = NULL ;
    gDefaults = NULL ;
    gEvents = NULL ;
    gThresholds = NULL ;
  }
  else
  {
    gDefaults->Release() ;gDefaults = NULL ;
    gSets->Release() ; gSets = NULL ;
    shutdown() ;
  }

#endif

//
// Ok lets create and open a new config
//
#ifdef DO_CREATE_CONFIGS_TEST 

  // Test making one of each version of config we support
  // *Note : If the close above fails, perhaps the file was read only - then this routine will put the root attribute
  // and other contents from the last opened config
  if( hr != S_OK ) shutdown() ;

  printf("\n*\n* Create new configuration file called: 'NewConfigDefaultV5.txt'\n*\n" ) ;
  if( gConfig->Create(L"c:\\NewConfigDefaultV5.txt", VARIANT_TRUE) != S_OK )
  {
    printf("** Error creating new configuration file\n" ) ; shutdown() ;
  }
  gConfig->Close() ;

  /*
  printf("\n*\n* Create new configuration file called: 'NewConfigV3.txt'\n*\n" ) ;
  gConfig->put_Version(3) ;
  if( gConfig->Create(L"c:\\NewConfigV3.txt", VARIANT_TRUE) != S_OK )
  {
    printf("** Error creating new configuration file\n" ) ; shutdown() ;
  }
  GetCollectionInterfaces() ;

  AddSets() ; ShowAllSets() ;
  AddParameters() ; ShowAllParameters() ;

  // Make sure to save out individual collections to the configuration
  if( gDefaultsChanged == true ) gDefaults->SaveTable() ;
  if( gSetsChanged == true ) gSets->SaveTable() ;

  hr = gConfig->Close() ;
  if( hr != S_OK ) shutdown() ;

  printf("\n*\n* Create new configuration file called: 'NewConfigV4.txt'\n*\n" ) ;

  gConfig->put_Version(4) ;
  if( gConfig->Create(L"c:\\NewConfigV4.txt", VARIANT_TRUE) != S_OK )
  {
    printf("** Error creating new configuration file\n" ) ; shutdown() ;
  }
  GetCollectionInterfaces() ;

  AddSets() ; ShowAllSets() ;
  AddParameters() ; ShowAllParameters() ;

  // Make sure to save out individual collections to the configuration
  if( gDefaultsChanged == true ) gDefaults->SaveTable() ;
  if( gSetsChanged == true ) gSets->SaveTable() ;

  hr = gConfig->Close() ;
  if( hr != S_OK ) shutdown() ;
  
  printf("\n*\n* Create new configuration file called: 'NewConfigV5.txt'\n*\n" ) ;

  gConfig->put_Version(5) ;
  if( gConfig->Create(L"c:\\NewConfigV5.txt", VARIANT_TRUE) != S_OK )
  {
    printf("** Error creating new configuration file\n" ) ; shutdown() ;
  }
  GetCollectionInterfaces() ;

  AddSets() ; ShowAllSets() ;
  AddParameters() ; ShowAllParameters() ;

  // Make sure to save out individual collections to the configuration
  if( gDefaultsChanged == true ) gDefaults->SaveTable() ;
  if( gSetsChanged == true ) gSets->SaveTable() ;
*/

#endif

//*****************************************************************************
//
//                        Large Parameter Defaults Test
//
//*****************************************************************************
#ifdef DO_LARGE_PARAMETER_DEFAULTS_TEST

  printf("\n*\n* Create new configuration file called: 'NewLargeConfig.txt'\n*\n" ) ;
  if( gConfig->Create(L"NewLargeConfig.txt", VARIANT_TRUE) != S_OK )
  {
    printf("** Error creating new configuration file\n" ) ; shutdown() ;
  }

  GetCollectionInterfaces() ;
  ShowVersion() ;

  ShowAllSets() ;
  // Testing results with the insert sort
  // 1. 98 sec for 10K parameters adding ascending order
  // 2. 28 sec for 10L parameters adding in decending order.
  AddLargeParameters(10000) ;

  printf("Complete with add to in-memory container...\n") ;

  // Make sure to save out individual collections to the configuration
  if( gDefaultsChanged == true ) gDefaults->SaveTable() ;
  if( gSetsChanged == true ) gSets->SaveTable() ;

  printf("Complete with save to file...\n") ;

  printf("Start full stride DB search\n") ;
  SearchLargeParameters() ;


#endif


  hr = gConfig->Close(VARIANT_TRUE) ;
  if( hr != S_OK )
  {
    printf("** Close and Save failed...\n" ) ;
  }

  shutdown() ;
}

//
// GetCollectionInterfaces
//
void GetCollectionInterfaces(void)
{
  // Get the parameterDefaults collection ( note this is easier in VB!! we are C++ programmers living in a VB world)
  HRESULT hr = gConfig->get_ParameterDefaults(&gDefaults) ; // Make sure you release this before releasing the config object
  if( hr != S_OK || gDefaults == NULL )
  {
    printf("** Error: Getting the Parameter Defaults object\n") ;
    shutdown() ;
  }

  // Get the parameter Sets collection
  hr = gConfig->get_ParameterSets(&gSets) ; // Make sure you release this before releasing the config object
  if( hr != S_OK || gSets == NULL )
  {
    printf("** Error: Getting the Parameter Sets object\n") ;
    shutdown() ;
  }
}

//
// OpenMsgLog
// 
void OpenMsgLog(void)
{
  // Log messages - this gives fuller error message information
  printf("- Open the message log called: msglog.txt\n" ) ;
  HRESULT hr = gConfig->OpenMessageLog(L"msglog.txt") ;
  if( hr != S_OK )
  {
    printf("** Unable to open message log?\n" ) ;
    shutdown() ;
  }
}

//
// ShowVersionFromFile
// 
void ShowVersionFromFile(void)
{
  HRESULT hr = gConfig->get_VersionFromFile(L"pfConfig", &b1 ) ;
  if( hr == S_OK )
  {
    ::wcstombs(tmpBuf1, b1, cMaxString) ;
    printf("- Version From File: %s\n", tmpBuf1 ) ;
  }
}

//
// ShowVersion
//
void ShowVersion(void)
{
  int ver ;
  HRESULT hr = gConfig->get_Version(&ver) ;
  if( hr == S_OK )
  {
    printf("- Version From Open: %d\n", ver ) ;
  }
  else
  {
    printf("- Unknown Version ?\n" ) ;
  }
}

//
// ShowAllSets
// 
void ShowAllSets(void)
{
  printf("- Show All Sets\n" ) ;

  long ns ;
  HRESULT hr = gSets->get_Count(&ns) ;
  if( hr != S_OK )
  {
    printf("** Error: Getting the parameter sets count\n") ; shutdown() ;
  }
  printf("- Total sets found: ( %d )\n", ns ) ;
 
  // Now loop through and show the parameter goodies
  VARIANT v; v.vt = VT_I4 ;
  for(int nn=1; nn <= ns; nn++ )
  {
    IParameterSet* p = NULL ;
    v.intVal = nn ;
    HRESULT hr = gSets->Item(v, &p) ;    
    if( hr != S_OK || p == NULL )
    {
      printf("** Error: Getting a parameter set item\n") ; shutdown() ;
    }

    tmpBuf1[0] = '\0' ; tmpBuf2[0] = '\0' ; tmpBuf3[0] = '\0' ;
    VARIANT_BOOL vb ;

    p->get_SetName(&b1) ; ::wcstombs(tmpBuf1, b1, cMaxString) ;
    p->get_Group(&b2) ; ::wcstombs(tmpBuf2, b2, cMaxString) ;
    p->get_IsActive(&vb) ; vb == -1 ? strcpy(tmpBuf3, "True") : strcpy(tmpBuf3, "False") ;

    printf( "\t%d. %s,%s,%s\n", nn, tmpBuf1, tmpBuf2, tmpBuf3 ) ;
    p->Release() ; // ** Note: Make sure you release because Item() AddRefed
  }
}

//
// AddNewSets
// 
void AddSets(void) 
{
  printf("- Add 2 Sets: Avionics and Structures\n" ) ;

  IParameterSet* p ;
  HRESULT hr = gSets->Add( L"Avionics", L"CNI", VARIANT_TRUE, &p )  ;
  if( hr != S_OK || p == NULL )
  {
    printf("** Error: Adding Set\n") ; shutdown() ;
  }
  p->put_Group(L"Weapons") ; // Change the group 
  p->Release() ; // ** Note: Make sure you release because Add() AddRefed

  hr = gSets->Add( L"Structures", L"", VARIANT_FALSE, &p )  ;
  if( hr != S_OK || p == NULL )
  {
    printf("** Error: Adding Set\n") ; shutdown() ;
  }
  p->Release() ; // ** Note: Make sure you release because Add() AddRefed
  gSetsChanged = true ;
}

//
// ShowAllAParameters
// 
void ShowAllParameters(void)
{
  printf("- Show All Parameters\n" ) ;

  // Get the total number of parameters in the config's ParameterDefaults Table
  long np ;
  HRESULT hr = gDefaults->get_Count(&np) ;
  if( hr != S_OK )
  {
    printf("** Error: Getting the parameter count\n") ; shutdown() ;
  }
  printf("- Total parameters found: ( %d )\n", np ) ;
 
  // Now loop through and show the parameter goodies
// TableDefinition ParameterDefaults|Parameter|ParamType|ParamGroup|ParamSubGroup|ShortName|LongName|Units|Color|Width|DataSourceType|DataSourceArguement|UpdateRate|LLNegative|LLPositive|TScaleRangeMin|TScaleRangeMax|TScaleAuto|FScaleRangeMin|FScaleRangeMax|FScaleAuto|WThreshRangeMin|WThreshRangeMax|WThreshColor|WThreshLabel|WThreshLineWidth|AThreshRangeMin|AThreshRangeMax|AThreshColor|AThreshLabel|AThreshLineWidth|FilterActive|FilterAlgorithm|FilterPassType|FilterLowCutoff|FilterHighCutoff|FilterOrder|WildPointRangeMin|WildPointRangeMax|WildPointCorMethod|WildPointCorValue|SignChange|NullCorrection|NullBaseline|NullAverage|NullBias|NullGroup|SpikeDetectionMethod|SpikeCorMethod|SpikeChangeLimit|TriggerCondition|AutoAnalysisType|ActivationOption|AutoBlockSize|PercentReduction|AutoFreqCalc|Frequency|UserFactor|ComputeType|ExcitationSignal|WindowType|Alpha|AveragingMethod|Overlap|BlocksPerAverage|BlockSize|StringLookupTable|UniqueIdStr|PropertyBag
  printf("ParameterDefaults,Parameter,ParamType,ParamGroup,ParamSubGroup,ShortName,LongName,Units\n" ) ;
  VARIANT v; v.vt = VT_I4 ;
  for(int n=1; n <= np; n++ )
  {
    IParameterDefault* p = NULL ;
    v.intVal = n ;
    HRESULT hr = gDefaults->Item(v, &p) ;    
    if( hr != S_OK || p == NULL )
    {
      printf("** Error: Getting a parameter item\n") ; shutdown() ;
    }

    tmpBuf1[0] = '\0' ; tmpBuf2[0] = '\0' ; tmpBuf3[0] = '\0'; tmpBuf4[0] = '\0' ;
    double sr ;

//    p->get_SetName(&b1) ;      
    p->get_Name(&b2) ;
    p->get_SampleRate(&sr) ;
    p->get_Units(&b3) ;

    IadsDataType type ;
    p->get_DataType(&type) ;
    if( type == iadsFloatingPoint ) strcpy( tmpBuf3,"float") ;
    else if( type == iadsDiscrete ) strcpy(tmpBuf2, "discrete") ;
    else if( type == iadsInteger) strcpy(tmpBuf3, "int") ;
    else strcpy(tmpBuf3, "undefined") ;

    ::wcstombs(tmpBuf1, b1, cMaxString) ;
    ::wcstombs(tmpBuf2, b2, cMaxString) ;
    ::wcstombs(tmpBuf4, b3, cMaxString) ;

    printf( "\t%d. %s,%s,%s,%f, %s\n", n, tmpBuf1, tmpBuf2, tmpBuf3, sr, tmpBuf4 ) ;
    p->Release() ; // ** Note: Make sure you release because Item() AddRefed
  }
}

//
// AddLargeParameters
//
void AddLargeParameters(int maxParms)
{
  printf("- Add (%d) Parameters with increasing numbers on base of AW0001X\n", maxParms ) ;

  wchar_t szBuffer[1024];
  IParameterDefault* p  = 0;
  HRESULT hr;

  for(int n=0; n < maxParms; n++ )
  {
    _snwprintf(szBuffer, sizeof(szBuffer), L"AW0001X%d", maxParms - n);
    
    VARIANT vtIndex;
    vtIndex.vt = VT_BSTR;
    vtIndex.bstrVal = ::SysAllocString(szBuffer);
    p = 0;

    //in our world, we lookup the item to see if it exists first
    //if it does, we just modify it
    //this lookup is what's chewing up load
    HRESULT retVal = gDefaults->Item(vtIndex, &p);
 
    if (!p )
    {
      hr = gDefaults->Add(szBuffer, iadsFloatingPoint, 10.0, &p )  ;
      if( hr != S_OK || p == NULL )
      {
          printf("** Error: Adding Parmeter\n") ; shutdown() ;
      }
      p->put_Group(L"Flutter") ;
      p->put_SubGroup(L"Wing") ;
      p->put_ShortName(L"Some Short Name") ;
      p->put_LongName(L"Some name longer then the short Name") ;
      p->put_Units(L"Deg's") ;
      p->put_DataSourceType(iadsTpp);

      p->Release() ; // ** Note: Make sure you release because Item() AddRefed
    }

    ::SysFreeString(vtIndex.bstrVal) ;
  }

  gDefaultsChanged = true ;
}

//
// SearchLargeParameters
//
void SearchLargeParameters(void)
{
  printf("- Searching for several parameters deep into the large parameter set\n" ) ;

  wchar_t szBuffer[1024];
  IParameterDefault* p  = NULL ;

  VARIANT vtIndex;
  vtIndex.vt = VT_BSTR;
 
  for(int n=0; n < 4; n++ )
  {
    if (n == 0 )
      _snwprintf(szBuffer, sizeof(szBuffer), L"AW0001X0", n);
    else if( n == 1 )
      _snwprintf(szBuffer, sizeof(szBuffer), L"AW0001X2000" );
    else if (n == 2 )
      _snwprintf(szBuffer, sizeof(szBuffer), L"AW0001X5000" );
    else if ( n == 3 )
      _snwprintf(szBuffer, sizeof(szBuffer), L"AW0001X9000" );
      
    vtIndex.bstrVal = ::SysAllocString(szBuffer);
                                                               
    HRESULT retVal = gDefaults->Item(vtIndex, &p);
    char str[1000] ;
    if( retVal != E_FAIL )
    {
      BSTR b ;
      p->get_Name(&b) ;
      ::wcstombs( str, b, 1000 ) ;
      printf(" Found: %s\n", str ) ; 

    }

    ::SysFreeString(vtIndex.bstrVal) ;

  }

}


//
// AddParameters
//
void AddParameters(void)
{
  printf("- Add 2 Parameters: AW0005X and AW0009X to the Avionics set\n" ) ;

  IParameterDefault* p ;
  HRESULT hr = gDefaults->Add( L"AW0005X", iadsFloatingPoint, 10.0, &p )  ;
  if( hr != S_OK || p == NULL )
  {
    printf("** Error: Adding Parmeter\n") ; shutdown() ;
  }
  p->put_Group(L"Flutter") ;
  p->put_SubGroup(L"Wing") ;
  p->put_ShortName(L"Some Short Name") ;
  p->put_LongName(L"Some name longer then the short Name") ;
  p->put_Units(L"Deg's") ;
  p->put_TimeScaleAuto(iadsOff) ;
//  p->put_DataSourceType(iadsDerived) ;
  p->put_DataSourceType(iadsIap) ;
  p->put_WarningThreshRangeMin(-10.0) ;
  p->put_WarningThreshRangeMax(10.0) ;
  p->put_WindowType( iadsFlatTop ) ;


  // Test the replacement character technique
  p->put_DataSourceArgument(L"Bit_Pick(1553L_Pilot.05.XT.17.19,1,1)||Bit_Pick(1553R_Co-Pilot.06.XT.17.19,2,2") ;

  BSTR eq ;
  p->get_DataSourceArgument(&eq) ;
  // End Test

  // Let's look at it
  BSTR b ;
  p->get_DataSourceArgument(&b) ;

  IadsDataSourceType dst ;
  p->get_DataSourceType(&dst) ;

  p->Release() ; // ** Note: Make sure you release because Item() AddRefed

  hr = gDefaults->Add( L"AW0009X", iadsInteger, 20.0, &p )  ;
  if( hr != S_OK || p == NULL )
  {
    printf("** Error: Adding Parmeter\n") ; shutdown() ;
  }
  p->put_Group(L"Flutter") ;
  p->put_SubGroup(L"Wing") ;
  p->put_ShortName(L"Some Short Name") ;
  p->put_LongName(L"Some name longer then the short Name") ;
  p->put_Units(L"Deg") ;

  p->Release() ; // ** Note: Make sure you release because Item() AddRefed


  hr = gDefaults->Add( L"AW0007X", iadsInteger, 20.0, &p )  ;
  if( hr != S_OK || p == NULL )
  {
    printf("** Error: Adding Parmeter\n") ; shutdown() ;
  }
  p->put_Group(L"Flutter") ;
  p->put_SubGroup(L"Wing") ;
  p->put_ShortName(L"Some Short Name") ;
  p->put_LongName(L"Some name longer then the short Name") ;
  p->put_Units(L"Deg") ;

  p->Release() ; // ** Note: Make sure you release because Item() AddRefed

  gDefaultsChanged = true ;
}

//
// RemoveParameters
//
void RemoveParameters(void)
{
  printf("- Remove Parameters: AW0009X\n" ) ;

  VARIANT vb; vb.bstrVal = ::SysAllocString(L"AW0009X") ; vb.vt = VT_BSTR ;
  HRESULT hr = gDefaults->Remove(vb) ;
  if( hr != S_OK ) 
  {
    int i = 10 ;
  }

  // This should fail
  ::SysFreeString(vb.bstrVal); vb.bstrVal =::SysAllocString(L"AW0009X") ;
  hr = gDefaults->Remove(vb) ;
  if( hr != S_OK ) 
  {
    int i = 10 ;
  }

  // Delete by index.
  VARIANT v; v.vt = VT_I4; v.intVal = 2 ;
  hr = gDefaults->Remove(v) ; 
  if( hr != S_OK ) 
  {
    int i = 10 ;
  }
  gDefaultsChanged = true ;
}


//**********************************************************************************//
//
//                                 Query Test Routines
//
//**********************************************************************************//

//
// DeleteQueryExamples

// Test Setup:
//  For this test make a copy of the Desktops table and rename it
// to "BogusDesktops".  THIS IS A DESTRUCTIVE TEST - so please use a copy !!
//
// Results:
//  These are the entries that should be remaining after the test.
//   1 Group|SubGroup, DesktopName = Desktop1, AnalysisWindowName = Window1|||
//   4 Loads|Strains|PestTestPrimary||
void DeleteQueryExamples(void)
{
  // Delete using where clause
  HRESULT hr = gConfig->Query( L"delete * from BogusDesktops where Group = 'Flutter'", NULL) ;
  if( hr == S_OK )
  {
    printf("Entry Deleted ...\n") ;
  }

  // Delete using "built-in" system id or row number
  hr = gConfig->Query( L"delete * from BogusDesktops where System.RowNumber = 3", NULL) ;
  if( hr == S_OK )
  {
    printf("Entry Deleted ...\n") ;
  }

  hr = gConfig->Save() ;
}


//
// CreateQueryExamples
//
// Test Setup:
//   None
//
// Results:
//   This function creates a new table called "BogusTable" and sets the header
// definition. The table will be at the end of the config file.  It uses various types
// in the header to demonstrate usasge.  List types head-up in a drop down in the IADS
// configuration tool.
//
void CreateQueryExamples(void)
{
  if( gConfig == NULL ) return ;

  VARIANT results ;
  HRESULT hr ;

  results.parray = NULL ;
  if( gConfig->Query( L"Select System.TableDefinition from ParameterAlias", &results ) != S_OK )
  {
//    HRESULT hr = gConfig->Query( L"create table BogusTable (Group String, X Int, Y Int, AutoScale list(True, False), Classification list(high,medium,low) )", NULL ) ;
    hr = gConfig->Query( L"create table ParameterAlias (IadsName String, DataSourceName string, PropertyBag string)", NULL ) ;
    if( hr == S_OK )
    {
      printf("Table Created ...\n") ;
    }
    hr = gConfig->Save() ;
  }
  else
  {
    printf("Table Already Exists...\n") ;
  }

}


//
// UpdateQueryExamples
//
// Test Setup:
//  For this test make a copy of the Desktops table and rename it
// to "BogusDesktops".  
//   
//
// Results:  your table should look similar to this.
//   1 Group|SubGroup, DesktopName = Desktop1, AnalysisWindowName = Window1|||
//   2 a|b|c|d|e
//   3 FQ|Pat|Pest Test||
//   4 Loads|Strains|PestTestPrimary||
//   5 a|b|c|d|e
//   6 a|b|c|d|e

void UpdateQueryExamples(void)
{
  // Use the update to append a new entry.
  HRESULT hr = gConfig->Query( L"update BogusDesktops set * = a|b|c|d|e", NULL ) ;
  if( hr == S_OK )
  {
    printf("Entry Added...\n") ;
  }

  // Modify all fields in the specified row(s)
  hr = gConfig->Query( L"update BogusDesktops set * = a|b|c|d|e where Group = 'Flutter'", NULL ) ;
  if( hr == S_OK )
  {
    printf("Entry Added...\n") ;
  }

  // Modify a single field in a specified row
  hr = gConfig->Query( L"update BogusDesktops set SubGroup = 'Pat' where Group = 'FQ'", NULL ) ;
  if( hr == S_OK )
  {
    printf("Entry Added...\n") ;
  }

  // This will perform a commit and a save - neccesary in order to see the changes 
  // in the file
  hr = gConfig->Save() ;
}


//
// SelectQueryExamples
//
// Test Setup:
//  None.
//
// Results:
//   Selected entries printed to the screen.
//
void SelectQueryExamples(void)
{
  VARIANT results ;

  // Get all fields for all entries in a particular table
  HRESULT hr = gConfig->Query( L"Select * from Desktops", &results ) ;
  if( hr == S_OK )
  {
    PrintQueryResults(&results) ;
    ::VariantClear(&results) ;
    printf("\n") ;
  }

  // Get all fields for selected entries using a "where" clause
  hr = gConfig->Query( L"Select * from Desktops where Group = 'Flutter'", &results ) ;
  if( hr == S_OK )
  {
    PrintQueryResults(&results) ;
    ::VariantClear(&results) ;
    printf("\n") ;
  }


  // Get all fields for selected entries using a compound "where" clause
  hr = gConfig->Query( L"Select * from Desktops where Group = 'Flutter' && AnalysisWindowName = 'DoubleIntTest'", &results ) ;
  if( hr == S_OK )
  {
    PrintQueryResults(&results) ;
    ::VariantClear(&results) ;
    printf("\n") ;
  }

  // Get one field for selected entries using a compound "where" clause
  hr = gConfig->Query( L"Select SubGroup from Desktops where Group = 'Flutter' && AnalysisWindowName = 'DoubleIntTest'", &results ) ;
  if( hr == S_OK )
  {
    PrintQueryResults(&results) ;
    ::VariantClear(&results) ;
    printf("\n") ;
  }

  // Get the unique row number for each entry
  hr = gConfig->Query( L"Select System.RowNumber from Desktops", &results ) ;
  if( hr == S_OK )
  {
    PrintQueryResults(&results) ;
    ::VariantClear(&results) ;
    printf("\n") ;
  }
}

//
// PrintQueryResults
//
void PrintQueryResults( VARIANT* v )
{
  BSTR* b ;
  SafeArrayAccessData( v->parray,(void**)&b ) ;
  int lBound = v->parray->rgsabound->lLbound ;
  int uBound = v->parray->rgsabound->cElements + lBound ;
  for( int i = lBound; i < uBound; i++ )
  {
    ::wcstombs( tmpBuf1, b[i], cMaxString ) ;
    printf("Result: %s\n", tmpBuf1 ) ;
  }
}

//
// ShowAllEvents
// 
void ShowAllEvents(void)
{
  printf("- Show All Events\n" ) ;

  // Get the total number of parameters in the config's ParameterDefaults Table
  long np ;
  HRESULT hr = gEvents->get_Count(&np) ;
  if( hr != S_OK )
  {
    printf("** Error: Getting the event count\n") ; shutdown() ;
  }
  printf("- Total events found: ( %d )\n", np ) ;
 
  // Now loop through and show the parameter goodies
// TableDefinition ParameterDefaults|Parameter|ParamType|ParamGroup|ParamSubGroup|ShortName|LongName|Units|Color|Width|DataSourceType|DataSourceArguement|UpdateRate|LLNegative|LLPositive|TScaleRangeMin|TScaleRangeMax|TScaleAuto|FScaleRangeMin|FScaleRangeMax|FScaleAuto|WThreshRangeMin|WThreshRangeMax|WThreshColor|WThreshLabel|WThreshLineWidth|AThreshRangeMin|AThreshRangeMax|AThreshColor|AThreshLabel|AThreshLineWidth|FilterActive|FilterAlgorithm|FilterPassType|FilterLowCutoff|FilterHighCutoff|FilterOrder|WildPointRangeMin|WildPointRangeMax|WildPointCorMethod|WildPointCorValue|SignChange|NullCorrection|NullBaseline|NullAverage|NullBias|NullGroup|SpikeDetectionMethod|SpikeCorMethod|SpikeChangeLimit|TriggerCondition|AutoAnalysisType|ActivationOption|AutoBlockSize|PercentReduction|AutoFreqCalc|Frequency|UserFactor|ComputeType|ExcitationSignal|WindowType|Alpha|AveragingMethod|Overlap|BlocksPerAverage|BlockSize|StringLookupTable|UniqueIdStr|PropertyBag
  printf("Group,SubGroup,User,IrigTime,Comment,PropBag\n" ) ;
  VARIANT v; v.vt = VT_I4 ;

  BSTR b1, b2, b3, b4, b5, b6 ;

  for(int n=1; n <= np; n++ )
  {
    IEvent* p = NULL ;
    v.intVal = n ;
    HRESULT hr = gEvents->Item(v, &p) ;    
    if( hr != S_OK || p == NULL )
    {
      printf("** Error: Getting a parameter item\n") ; shutdown() ;
    }

    tmpBuf1[0]=tmpBuf2[0]=tmpBuf3[0]=tmpBuf4[0]=tmpBuf5[0]=tmpBuf6[0] = '\0' ;

    p->get_Group(&b1) ;
    p->get_SubGroup(&b2) ;
    p->get_User(&b3) ;
    p->get_IrigTime(&b4) ;
    p->get_Name(&b5) ;
    p->get_PropBag(&b6) ;

    if( b1 != (BSTR)NULL )
      ::wcstombs( tmpBuf1, b1, cMaxString ) ;
    if( b2 != (BSTR)NULL )
      ::wcstombs( tmpBuf2, b2, cMaxString ) ;
    if( b3 != (BSTR)NULL )
      ::wcstombs( tmpBuf3, b3, cMaxString ) ;
    if( b4 != (BSTR)NULL )
      ::wcstombs( tmpBuf4, b4, cMaxString ) ;
    if( b5 != (BSTR)NULL )
      ::wcstombs( tmpBuf5, b5, cMaxString ) ;
    if( b6 != (BSTR)NULL )
      ::wcstombs( tmpBuf6, b6, cMaxString ) ;

    printf( "\t%d. %s,%s,%s,%s,%s,%s\n", n, tmpBuf1, tmpBuf2, tmpBuf3, tmpBuf4, tmpBuf5, tmpBuf6 ) ;
    p->Release() ; // ** Note: Make sure you release because Item() AddRefed
  }
}

//
// ShowAllThresholds
// 
void ShowAllThresholds(void)
{
  printf("- Show All Thresholdss\n" ) ;

  // Get the total number of parameters in the config's ParameterDefaults Table
  long np ;
  HRESULT hr = gThresholds->get_Count(&np) ;
  if( hr != S_OK )
  {
    printf("** Error: Getting the Threshold count\n") ; shutdown() ;
  }
  printf("- Total Thresholds found: ( %d )\n", np ) ;
 
// Group|SubGroup|User|ThesholdLevel|AnalysisWindow|DataDisplayType|Parameter|Time|Value|DataDisplayName|Comment|PropertyBag

  // Now loop through and show the parameter goodies
  printf("Group,SubGroup,User,IrigTime,Comment,PropBag\n" ) ;
  VARIANT v; v.vt = VT_I4 ;

  BSTR b1, b2, b3, b4, b5, b6 ;

  for(int n=1; n <= np; n++ )
  {
    IThreshold* t = NULL ;
    v.intVal = n ;
    HRESULT hr = gThresholds->Item(v, &t) ;    
    if( hr != S_OK || t == NULL )
    {
      printf("** Error: Getting a Threshold item\n") ; shutdown() ;
    }

    tmpBuf1[0]=tmpBuf2[0]=tmpBuf3[0]=tmpBuf4[0]=tmpBuf5[0]=tmpBuf6[0] = '\0' ;

    t->get_Group(&b1) ;
    t->get_SubGroup(&b2) ;
    t->get_User(&b3) ;
    t->get_IrigTimeAtBreak(&b4) ;
    t->get_ParameterName(&b5) ;
    t->get_PropBag(&b6) ;

    if( b1 != (BSTR)NULL )
      ::wcstombs( tmpBuf1, b1, cMaxString ) ;
    if( b2 != (BSTR)NULL )
      ::wcstombs( tmpBuf2, b2, cMaxString ) ;
    if( b3 != (BSTR)NULL )
      ::wcstombs( tmpBuf3, b3, cMaxString ) ;
    if( b4 != (BSTR)NULL )
      ::wcstombs( tmpBuf4, b4, cMaxString ) ;
    if( b5 != (BSTR)NULL )
      ::wcstombs( tmpBuf5, b5, cMaxString ) ;
    if( b6 != (BSTR)NULL )
      ::wcstombs( tmpBuf6, b6, cMaxString ) ;

    printf( "\t%d. %s,%s,%s,%s,%s,%s\n", n, tmpBuf1, tmpBuf2, tmpBuf3, tmpBuf4, tmpBuf5, tmpBuf6 ) ;
    t->Release() ; // ** Note: Make sure you release because Item() AddRefed
  }
}

//
// ShowAllTestpoints
//
//   TableDefinition Group|SubGroup|User|TestPoint|Description|Maneuver|StartTime|StopTime|PropertyBag
//   TableType String|String|String|listQuery(select Distinct TestPoint from PlannedTestPoints)|String|listQuery(select Distinct Maneuver from PlannedTestPoints)|String|String|String
// 
void ShowAllTestpoints(void)
{
  printf("- Show All TestPoints\n" ) ;

  // Get the total number of parameters in the config's ParameterDefaults Table
  long np ;
  HRESULT hr = gTestpoints->get_Count(&np) ;
  if( hr != S_OK )
  {
    printf("** Error: Getting the Testpoints count\n") ; shutdown() ;
  }
  printf("- Total Testpoints found: ( %d )\n", np ) ;
 

  // Now loop through and show the parameter goodies
  printf("Group,SubGroup,User,TestPoint,Description,Maneuver,StartTime,StopTime,PropBag\n" ) ;
  VARIANT v; v.vt = VT_I4 ;

//  BSTR b1, b2, b3, b4, b5, b6, b7, b8, b9 ;

  for(int n=1; n <= np; n++ )
  {
    ITestpoint* t = NULL ;
    v.intVal = n ;
    HRESULT hr = gTestpoints->Item(v, &t) ;    
    if( hr != S_OK || t == NULL )
    {
      printf("** Error: Getting a Testpoint item\n") ; shutdown() ;
    }

    tmpBuf1[0]=tmpBuf2[0]=tmpBuf3[0]=tmpBuf4[0]=tmpBuf5[0]=tmpBuf6[0]=tmpBuf7[0]=tmpBuf8[0]=tmpBuf9[0] = '\0' ;

    t->get_Group(&b1) ;
    t->get_SubGroup(&b2) ;
    t->get_User(&b3) ;
    t->get_Testpoint(&b4) ;
    t->get_Description(&b5) ;
    t->get_Maneuver(&b6) ;
    t->get_StartTime(&b7) ;
    t->get_StopTime(&b8) ;
    t->get_PropBag(&b9) ;

    if( b1 != (BSTR)NULL ) ::wcstombs( tmpBuf1, b1, cMaxString ) ;
    if( b2 != (BSTR)NULL ) ::wcstombs( tmpBuf2, b2, cMaxString ) ;
    if( b3 != (BSTR)NULL ) ::wcstombs( tmpBuf3, b3, cMaxString ) ;
    if( b4 != (BSTR)NULL ) ::wcstombs( tmpBuf4, b4, cMaxString ) ;
    if( b5 != (BSTR)NULL ) ::wcstombs( tmpBuf5, b5, cMaxString ) ;
    if( b6 != (BSTR)NULL ) ::wcstombs( tmpBuf6, b6, cMaxString ) ;
    if( b7 != (BSTR)NULL ) ::wcstombs( tmpBuf6, b7, cMaxString ) ;
    if( b8 != (BSTR)NULL ) ::wcstombs( tmpBuf6, b8, cMaxString ) ;
    if( b9 != (BSTR)NULL ) ::wcstombs( tmpBuf6, b9, cMaxString ) ;

    printf( "\t%d. %s,%s,%s,%s,%s,%s,%s,%s,%s\n", n,
            tmpBuf1, tmpBuf2, tmpBuf3, tmpBuf4, tmpBuf5,
            tmpBuf6, tmpBuf7, tmpBuf8, tmpBuf9 ) ;
    t->Release() ; // ** Note: Make sure you release because Item() AddRefed
  }
}

//
// AddTestpoints
//
void AddTestpoints(void)
{
  printf("- Add 2 TestPoints\n" ) ;

  ITestpoint* tp ;
  HRESULT hr = gTestpoints->Add( L"001-002-003", L"100:23:14:100", L"100:23:20:200", &tp )  ;
  if( hr != S_OK || tp == NULL )
  {
    printf("** Error: Adding Parmeter\n") ; shutdown() ;
  }
  tp->put_Group(L"Loads") ;
  tp->put_SubGroup(L"Tail") ;
  tp->put_User(L"Jim") ;
  tp->put_Description(L"2G Roll") ;
  tp->put_Maneuver(L"2G Roll Maneuver") ;

  tp->Release() ; // ** Note: Make sure you release because Item() AddRefed

  hr = gTestpoints->Add( L"002-003-004", L"100:23:22:100", L"100:23:50:200", &tp )  ;
  if( hr != S_OK || tp == NULL )
  {
    printf("** Error: Adding TestPoints\n") ; shutdown() ;
  }

  tp->put_Group(L"Loads") ;
  tp->put_SubGroup(L"Wing") ;
  tp->put_User(L"Marty") ;
  tp->put_Description(L"10G Roll") ;
  tp->put_Maneuver(L"10G Roll Maneuver") ;

  tp->Release() ;
}


//
// AddEvents
//
void AddEvents(void)
{
  printf("- Add 2 Events\n" ) ;

  IEvent* p ;
  HRESULT hr = gEvents->Add( L"E4", L"100:23:14:100", &p )  ;
  if( hr != S_OK || p == NULL )
  {
    printf("** Error: Adding Parmeter\n") ; shutdown() ;
  }
  p->put_Group(L"Loads") ;
  p->put_SubGroup(L"Tail") ;
  p->put_User(L"Ben") ;

  p->Release() ; // ** Note: Make sure you release because Item() AddRefed

  hr = gEvents->Add( L"E6", L"100:23:14:100", &p )  ;
  if( hr != S_OK || p == NULL )
  {
    printf("** Error: Adding Parmeter\n") ; shutdown() ;
  }

  p->put_Group(L"Loads") ;
  p->put_SubGroup(L"Wing") ;
  p->put_User(L"Marty") ;


  p->Release() ; // ** Note: Make sure you release because Item() AddRefed
}


//
//
//
void MergeParametersFromTwoConfigs()
{
  IIadsConfig* gConfigUser = NULL ;
  IIadsConfig* gConfigInstr = NULL ;
  HRESULT hr1 = ::CoCreateInstance( CLSID_IadsConfig, NULL, CLSCTX_ALL, IID_IIadsConfig, (void**)&gConfigUser ) ;
  HRESULT hr2 = ::CoCreateInstance( CLSID_IadsConfig, NULL, CLSCTX_ALL, IID_IIadsConfig, (void**)&gConfigInstr ) ;
  if( hr1 != S_OK || hr2 != S_OK || gConfigUser == NULL || gConfigInstr == NULL )
  {
    printf("** Unable to create the IadsConfig object ?\n" ) ;
    shutdown() ;
  }

  hr1 = gConfigUser->Open(L"configFromUsers", VARIANT_TRUE ) ;
  hr2 = gConfigInstr->Open(L"configFromInstr", VARIANT_TRUE ) ;
  if( hr1 != S_OK || hr2 != S_OK )
  {
    printf("** Unable to open configuration file\n" ) ;
    shutdown() ;
  }

//
// Delete all PCM parameters from the User's configuration
//
  hr1 = gConfigUser->Query( L"delete * from ParameterDefaults where ParamGroup = 'PCM' || ParamGroup = 'PCM_DERIVED'", NULL) ;
  if( hr1 != S_OK )
  {
    shutdown() ;
  }

//
// Get all the new PCM parameters from the instrumentation config
//
  VARIANT qr ;
  hr1 = gConfigInstr->Query( L"select * from ParameterDefaults", &qr) ;
  if( hr1 != S_OK )
  {
    shutdown() ;
  }

//
// Now update the "user" config with these
//
  const int cMaxString = 10000 ;
  char parmDef[cMaxString] ;
  char updQ[cMaxString] ;

  CComBSTR bQ ;

  BSTR* b ;
  SafeArrayAccessData( qr.parray,(void**)&b ) ;
  int lBound = qr.parray->rgsabound->lLbound ;
  int uBound = qr.parray->rgsabound->cElements + lBound ;
  for( register int n = lBound; n < uBound; n++ )
  {
    // this is a little hacky - but hey it's fast ...
    ::wcstombs( parmDef, b[n], cMaxString ) ;
    sprintf( updQ, "update ParameterDefaults set * = %s where 0", parmDef ) ;

    bQ = updQ ;
    hr1 = gConfigUser->Query( bQ, NULL ) ;

    if( !(n%2000) )
      printf("Updated ( %d ) parameters in User config...\n", n ) ;
  }

  hr1 = gConfigUser->Save() ;

//
// Shutdown
//
  hr1 = gConfigUser->Close(VARIANT_TRUE) ;
  hr2 = gConfigInstr->Close(VARIANT_TRUE) ;

  shutdown() ;
}